/***********************************************************************
    gutenfetch - query and fetch electronic texts from project gutenberg
    Copyright (C) 2001, 2002, 2003 Russell Francis 

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the 
	
	Free Software Foundation, Inc.,
	59 Temple Place, Suite 330,
	Boston, MA 02111-1307
	USA

$Id: gutenfetch.c,v 1.40 2004/07/11 19:05:54 johntabularasa Exp $
***********************************************************************/

#include "stddefs.h"
#ifdef HAVE_STDIO_H
#	include <stdio.h>
#endif
#ifdef HAVE_STDLIB_H
#	include <stdlib.h>
#endif 
#ifdef ENABLE_NLS
#	include <locale.h>
#	include <libintl.h>
#endif
#if (HAVE_GETOPT_H && (USE_C_GETOPT_LONG == 1))
#	include <getopt.h>
#else
#	include "getopt_long.h"
#endif
#ifdef HAVE_SYS_UIO_H
#	include <sys/uio.h>
#endif
#ifdef HAVE_SYS_TYPES_H
#	include <sys/types.h>
#endif
#ifdef HAVE_SYS_STAT_H
#	include <sys/stat.h>
#endif
#if (HAVE_ASSERT_H == 1)
#	include <assert.h>
#endif
#if (HAVE_STRING_H == 1)
#	include <string.h>
#endif
#if (HAVE_GUTENFETCH_H == 1)
#	include <gutenfetch.h>
#endif
#include "opt.h"

int main(int, char **);
static int parse_options(int *, char ***);
static void print_usage(void);
static void print_version(void);

static void list_books(void);
static void fetch_book(void);

int
display_progress(void *, double, double, double, const char *);

int
display_progress(
	void *data,
	double percent_complete,
	double total,
	double items,
	const char *desc)
{
	int i;
	char pbar[51];
	static double last_percent_complete = 1.0;

	if( opt_get_verbose() == TRUE )
	{
		if((last_percent_complete > percent_complete)) {
			fprintf( stderr, "\n%s\n", desc );
		} else {
			for( i = 0; i < 50; ++i ) {
				if( (double)((double)i / 50.0) < percent_complete ) {
					pbar[i] = 'O';
				} else {
					pbar[i] = '.';
				}
			}
			pbar[50] = '\0';
			fprintf( stderr, "[%s] %1.1f/%1.1f\r", pbar, items, total );	
		}
		last_percent_complete = percent_complete;
	}
	return( 0 );
}
	

/**
 * Main entry point for gutenfetch
 *
 * @param argc The number of command line arguments passed.
 * @param argv an array of NULL terminated strings for each argument.
 * @return 1 if all is ok, 0 if not.
 */
int main(int argc, char **argv)
{
	gutenfetch_error_t gutenerr;

#ifdef ENABLE_NLS
	/* 
	 * setup i8ln with gettext 
	 */
	setlocale (LC_ALL, "");
	bindtextdomain (PACKAGE, LOCALEDIR);
	textdomain (PACKAGE);
#endif

	gutenerr = gutenfetch_init();
	if( gutenerr != GUTENFETCH_OK ) {
		fprintf( stderr, "ERROR: %s\n", 
			gutenfetch_error_to_string( gutenerr ) );
		gutenfetch_shutdown();
		exit( 1 );
	}

	opt_init(); 
	if(parse_options(&argc, &argv)) {
		opt_shutdown();	
		gutenfetch_shutdown();
		exit (1);
	}

	/* Enable/Disable caching. */
	gutenfetch_cache_enable( opt_get_cache() );

	/*
	 * load config file get default servers etc...
	 */
	if (opt_load_config() == FALSE) {
		opt_generate_new_config_file();
	}

	switch (opt_get_action()) {
	case ACTION_NONE:
		fprintf(stderr, _("!!! You must specify an action to take with either the -m, -l or -f flags.\n"));
	case ACTION_PRINT_USAGE:
		print_usage();	
		break;
	case ACTION_PRINT_VERSION:
		print_version();
		break;
	case ACTION_MANUAL_SERVER_SELECT:
		opt_generate_new_config_file();
		break;
	case ACTION_LIST_BOOKS:
		list_books();
		break;
	case ACTION_FETCH_BOOK:
		fetch_book();
		break;
	case ACTION_ERROR:	
	default:
		fprintf(stderr, _("Action error\n"));
		break;
	}

	/* Release all resources. */
	opt_shutdown();
	gutenfetch_shutdown();

	return 1;
}


/**
 * Parse the command line options.
 *
 * Given a pointer to argc and argv, parse the command line options.
 *
 * @param argc A pointer to the argc parameter passed to main().
 * @param argv A pointer to the argv parameter passed to main().
 * @return Whether we should continue or quit on return.
 */
int
parse_options(
	int *argc,
	char ***argv)
{
	extern char *optarg;
	extern int optind;
	static struct option longopts[] = {
		{ "help",		no_argument,		NULL, 	'h' },
		{ "version",	no_argument,		NULL,	'V' },
		{ "verbose", 	no_argument, 		NULL,	'v' },
		{ "quite", 	no_argument, 		NULL,	'q' },
		{ "list",	no_argument,		NULL,	'l' },
		{ "make-config",no_argument,		NULL,	'm' },
		{ "output",	required_argument,	NULL,   'o' },
		{ "server", 	required_argument, 	NULL,	's' },
		{ "fetch",	required_argument,	NULL,	'f' },
		{ "filter",	no_argument,		NULL,	'F' },
#ifdef DEBUG		
		{ "debug",	no_argument,		NULL,	'D' },
#endif /* DEBUG */		
		{ "clear-filter",no_argument,		NULL,	'C' },
		{ "enable-cache",no_argument,		NULL,	'c' },
		{ "disable-cache",no_argument,		NULL,	'd' },
		{ 0, 0, 0, 0 }
	};
	int ch, retval = FALSE;

	while (
	((ch = getopt_long( *argc, *argv, "DdCcFmhvVqlo:s:f:",
	longopts, NULL)) != -1) && (retval != TRUE)) {
		switch (ch) {
		case 'm':   // generate a new config file.
			opt_set_action(ACTION_MANUAL_SERVER_SELECT);
			break;
		case 'V':	// version message
			opt_set_action(ACTION_PRINT_VERSION);
			break;
		case 'v':	// be verbose
			opt_set_verbose(TRUE);
			break;
		case 'q':	// be quite
			opt_set_verbose(FALSE);
			break;
		case 'o':	// output file
			opt_set_output_file(optarg);
			break;
		case 's':	// force server
			gutenfetch_set_active_server(optarg);
			break;
		case 'F':	// filter the results
			opt_set_filter(TRUE);
			break;
		case 'C':	// do not filter the results
			opt_set_filter(FALSE);
			break;
		case 'f':	// fetch a book
			opt_set_action(ACTION_FETCH_BOOK);
			if (opt_set_bookbase(optarg) == FALSE)
				retval = TRUE;
			break;
		case 'l':	// list all books on server
			opt_set_action(ACTION_LIST_BOOKS);
			break;
		case 'c':
			opt_set_cache( TRUE );
			break;
		case 'd':
			opt_set_cache( FALSE );
			break;
#if defined(DEBUG)			
		case 'D':   // set debugging flag
			opt_set_debug(TRUE);
			break;
#endif /* DEBUG */			
		case 'h':	// help message
		case '?':	
		default:
			opt_set_action(ACTION_PRINT_USAGE);
			break;
		}
	}

	*argc -= optind;
	*argv += optind;

	return retval;
}

/**
 * Print the version of gutenfetch.
 *
 * This displays the current version of gutenfetch and libraries
 * which it is compiled against.
 */
void
print_version(void)
{
	fprintf(stderr, "\n%s by Russell Francis\n", PACKAGE_STRING );
	fprintf(stderr, _("send bug reports to Russ at <%s>\n"), PACKAGE_BUGREPORT);
	fprintf(stderr, _("Licensed under the Gnu Public License (GPL)\n\n"));
	fprintf(stderr, _("%s is compiled against the following libraries.\n"), PACKAGE);
	fprintf(stderr, "libgutenfetch - %s\n\n", gutenfetch_version() );
}


/**
 * Print the usage instructions for gutenfetch.
 *
 */
void print_usage(void)
{
	print_version();
	fprintf(stderr, _("usage: %s [options] [-m|-l|-f [bookid]] [-o file]\n\n"),
		PACKAGE); 
	fprintf(stderr, "[--help,         -h] -\t%s\n",
		_("Display this help screen."));
#if defined(DEBUG)
	fprintf(stderr, "[--debug,        -D] -\t%s\n",
		_("Display debugging information."));
#endif /* DEBUG */		
	fprintf(stderr, "[--version,      -V] -\t%s\n",
		_("Display the version."));
	fprintf(stderr, "[--make-config,  -m] -\t%s\n",
		_("Generate a new .gutenfetchrc file."));
	fprintf(stderr, "[--verbose,      -v] -\t%s\n",
		_("Display verbose messages."));
	fprintf(stderr, "[--quiet,        -q] -\t%s\n",
		_("Display no messages to the user."));
	fprintf(stderr, "[--list,         -l] -\t%s\n",
		_("List all books on the Project Gutenberg servers."));
	fprintf(stderr, "[--output=file,  -o] -\t%s\n",
		_("Specify a file to write the output to."));	
	fprintf(stderr, "[--server=serv,  -s] -\t%s\n",
		_("Force the use of a particular server."));
	fprintf(stderr, "[--fetch=bookid, -f] -\t%s\n",
		_("Fetch a particular book off of the server."));
	fprintf(stderr, "[--filter,       -F] -\t%s\n",
		_("Filter the results. (default)"));
	fprintf(stderr, "[--clear-filter, -C] -\t%s\n",
		_("When listing books, do not filter the results."));
	fprintf(stderr, "[--enable-cache, -c] -\t%s\n",
		_("Enable caching of retrieved etexts. (default)"));
	fprintf(stderr, "[--disable-cache,-d] -\t%s\n",
		_("Disable caching of retrieved etexts."));
	fprintf(stderr, "\n");
}

void
fetch_book(void)
{
	FILE *fp;
	const char *output_file;
	gutenfetch_error_t gutenerr;
	gutenfetch_etext_t **etexts = NULL;
	gutenfetch_file_t file;
	int book_id;
	int i, j;

	gutenerr = gutenfetch_get_listing(
		&etexts,
		LIST_ALL,
		display_progress,
		NULL );

	if( gutenerr != GUTENFETCH_OK ) {
		fprintf(stderr, "Error fetching books: %s\n", 
			gutenfetch_error_to_string( gutenerr ) );

		return;
	}

	/* open the output file. */
	output_file = opt_get_output_file();
	if( (output_file == NULL) || (strcmp(output_file, "-") == 0 ) ) {
		fp = fdopen(1, "w");
	} else {
		fp = fopen( output_file, "w");
	}
	
	if( fp == NULL ) {
		if( opt_get_verbose() == TRUE ) {
			fprintf(stderr, _("Unable to open output file '%s'.\n"), output_file );
		}
	} else {


		book_id = strtol(opt_get_bookbase(), NULL, 10);

		for(i = 0; (etexts[i] != NULL); ++i ) {
			if( etexts[i]->id == book_id ) {
				gutenfetch_detail_etext( etexts[i] );
				if( etexts[i]->entry != NULL ) {
					for(j = 0; etexts[i]->entry[j] != NULL; ++j ) {
						if( etexts[i]->entry[j]->format.plain_text == 1) {
							gutenerr = gutenfetch_get_etext(
								etexts[i]->entry[j],
								&file,
								FALSE,
								display_progress,
								NULL);

							if( gutenerr == GUTENFETCH_OK ) {
								if( file.contents != NULL ) {
								#ifndef WIN32
									gutenfetch_ms_strip_text_buffer( file.contents );
								#endif
									fprintf( fp, "%s\n", file.contents );
									FREE_NULL( file.contents );
								}
							} else {
								fprintf( stderr, "Error: %s\n", gutenfetch_error_to_string( gutenerr ) );
							}
							FREE_NULL( file.filename );
							break;
						} 
					}
					if( etexts[i]->entry[j] == NULL ) {
						fprintf( stderr, "Unable to find an etext entry which is of type plain text.\n" );
					}
				}
				break;
			}
		}
		fclose(fp); 
	}
}

void
list_books(void)
{
	gutenfetch_error_t gutenerr;
	gutenfetch_etext_t **etexts = NULL;
	unsigned int i;
	const char *output_file;
	FILE *fp;
	char *buffer = NULL;

	if( opt_get_filter() == TRUE ) {
		gutenerr = gutenfetch_get_listing(
			&etexts,
			LIST_ALL,
			display_progress,
			NULL );
	} else {
		gutenerr = gutenfetch_get_raw_listing( &buffer,
			LIST_ALL,
			NULL,
			NULL);
	}

	if( gutenerr != GUTENFETCH_OK ) {
		fprintf(stderr, "Error fetching books: %s\n", 
			gutenfetch_error_to_string( gutenerr ) );
		if( opt_get_filter() == FALSE ) {
			FREE_NULL(buffer);
		}
		return;
	}

	/* open the output file. */
	output_file = opt_get_output_file();
	if( (output_file == NULL) || (strcmp(output_file, "-") == 0 ) ) {
		fp = fdopen(1, "w");
	} else {
		fp = fopen( output_file, "w");
	}
	
	if( fp == NULL ) {
		if( opt_get_verbose() == TRUE ) {
			fprintf(stderr, _("Unable to open output file '%s'.\n"), output_file );
		}
	} else {

		if( opt_get_filter() == FALSE ) {
			fprintf( fp, "%s", buffer );
		} else {
			/* list all books. */
			for( i = 0; etexts[i] != NULL; ++i ) {
				fprintf( fp, "%s\n", etexts[i]->full );
			}
		}

		fclose(fp);
	}

	if( opt_get_filter() == FALSE ) {
		FREE_NULL(buffer);
	}
}
