/***********************************************************************
	gutenfetch - query and fetch electronic texts from project gutenberg
    Copyright (C) 2001, 2002, 2003 Russell Francis 

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
	USA

	Last updated on $Date: 2004/07/11 19:05:54 $ by $Author: johntabularasa $.
***********************************************************************/
#include "stddefs.h"
#ifdef HAVE_ASSERT_H
#	include <assert.h>
#endif
#ifdef HAVE_STDIO_H
#	include <stdio.h>
#endif
#ifdef HAVE_STDLIB_H
#	include <stdlib.h>
#endif
#ifdef HAVE_STRING_H
#	include <string.h>
#endif
#ifdef HAVE_STRINGS_H
#	include <strings.h>
#endif
#ifdef HAVE_SYS_TYPES_H
#	include <sys/types.h>
#endif
#ifdef HAVE_SYS_STAT_H
#	include <sys/stat.h>
#endif
#ifdef HAVE_SYS_UIO_H
#	include <sys/uio.h>
#endif
#ifdef HAVE_GUTENFETCH_H
#	include <gutenfetch.h>
#endif
#include "opt.h"
#include "util.h"

#ifdef WIN32
#	define CONFIG_FILE "gutenfetch.rc"
#else
#	define CONFIG_FILE ".gutenfetchrc"
#endif

#if defined(DEBUG)
	static int debug = FALSE;
#endif

static int verbose = TRUE;	/* whether we should be verbose or not */
static char *output_file = NULL;
static char *bookbase = NULL;
static int filter = TRUE; 			// should we filter the book list
						// before displaying
static int enable_cache = TRUE;			// Should libgutenfetch enable cacheing?
static action_t action = ACTION_NONE;

/* Local Function Definitions */

static int opt_try_load_file(const char *);
static gutenfetch_server_t * opt_get_server_selection(gutenfetch_server_t **);
static gutenfetch_server_t ** opt_get_continent_selection(void);


#if defined(DEBUG)
/**
 * Set get whether debugging information should be displayed or not?
 *
 * @return TRUE if we should display debugging information, FALSE
 * if we should not return debugging information.
 */
int
opt_get_debug(void)
{
	return debug;
}


/**
 * Set the whether or not we should display debugging information.
 *
 * @param d TRUE to display debugging information, FALSE to not 
 * display it.
 */
void
opt_set_debug(const int d)
{
	debug = d;
}	
#endif


/**
 * Initialize anything from the opt.c file that needs it.
 *
 */
int
opt_init(void)
{
	DEBUG_MSG0("*** opt_init()\n");
	output_file = strdup("-");
	return TRUE;
}


/**
 * Shutdown all resources used and held by the opt.c file.
 *
 */
void
opt_shutdown(void)
{
	DEBUG_MSG0("*** opt_shutdown()\n");
	FREE_NULL(bookbase);
	FREE_NULL(output_file);
}


/**
 * Load gutenfetch configuration from a file.
 *
 * This function cycles through all possible locations for
 * a configuration file and stops at the first one it finds.
 * currently, it looks for $HOME/.gutenfetchrc then /etc/gutenfetchrc
 * If either of those files are found it uses them to load the 
 * default servers, and other options.
 *
 * @return The return value is TRUE if a server was found and
 *         we were able to load the configuration information
 *         without error.  FALSE is returned if we cannot find
 *         a file or an error was encountered while loading it.
 */
int
opt_load_config(void)
{
	char *home_dir = util_gethomedir();
	char *filename = NULL;
	int found_config_file = FALSE;

	DEBUG_MSG0("*** opt_load_config()\n");
	DEBUG_MSG1("*** Your home directory was found to be: %s\n", home_dir);

	/* find $HOME/.gutenfetchrc */
	if (home_dir != NULL) {
		filename = util_strcat(home_dir, 
			DIR_SEPARATOR, CONFIG_FILE, NULL);

		if(opt_try_load_file(filename))
			found_config_file = TRUE;
		FREE_NULL(filename);	
	}
	
#ifndef WIN32	 
	/* find /etc/gutenfetchrc */
	/* if we are not on windows ... */
	if (!found_config_file) {
		filename = strdup("/etc/gutenfetchrc");
		if (opt_try_load_file(filename))
			found_config_file = TRUE;
		FREE_NULL(filename);	
	}
#endif
	return found_config_file;
}


/**
 * Write the current configuration to a file.
 *
 * The configuration will be written to the file
 * $HOME/.gutenfetchrc  If you have root priviliges
 * you may manually copy this file to /etc/gutenfetchrc
 *
 */
int
opt_write_config(void)
{
	int retval = FALSE;
	char *home_dir = util_gethomedir();
	char *filename;
	FILE *fp;
	gutenfetch_server_t *gserver;

	DEBUG_MSG0("*** opt_write_config()\n");

	if (home_dir != NULL) {
		DEBUG_MSG1("*** opt_write_config() - home_dir = '%s'\n", home_dir );
		filename = util_strcat(home_dir, DIR_SEPARATOR, 
			CONFIG_FILE, NULL);
		assert(filename != NULL);
	
		DEBUG_MSG1("*** opt_write_config() - filename = '%s'\n", filename );	
		fp = fopen(filename, "w");
		if (fp == NULL) {
			if (verbose) {
				fprintf(stderr, _("Unable to open the configuration file '%s'.\n"),
					filename);
			}
			retval = FALSE;
		} else {
			gserver = gutenfetch_get_active_server();
			if( gserver == NULL ) {
				fprintf(stderr, _("Unable to get the active server.\n") );
			} else {
				if( gserver->host == NULL ) {
					fprintf(stderr, _("Unable to determin hostname of server.\n") );
				} else {
					DEBUG_MSG1( "*** opt_write_config() - server = '%s'\n", gserver->host );
					fprintf( fp, "server %s\n", gserver->host );
				}
				gutenfetch_free_server( gserver );
			}

			fprintf(fp, "verbose %s\n", verbose ? "true" : "false");

			if (fclose(fp) != 0) {
				if (verbose) {
					fprintf(stderr, _("Unable to close file '%s'."), filename);
				}
			}	
		}

		FREE_NULL(filename);
	}
	return retval;
}



/**
 * This will try to open the given filename and set the
 * configuration from its contents.
 *
 * @param filename The file to open and read as a configuration
 *        file.
 * @return Return TRUE if we loaded the configuration successfully.
 *         FALSE if we were unable to load the configuration.
 */
int
opt_try_load_file(const char *filename)
{
	FILE *fp;
	char *line = NULL;
	char *action = NULL;
	char *value = NULL;
	//char *new_value = NULL;
	unsigned int line_number = 0;
	gutenfetch_error_t gutenerr;
	assert (filename != NULL);

	DEBUG_MSG1("*** opt_try_load_file('%s')\n", filename);
	
	fp = fopen(filename, "r");
	if (fp == NULL) {
		if (verbose) {
			fprintf(stderr, _("Unable to open the configuration file '%s'.\n"),
				filename);
		}
		return FALSE;	
	}
	DEBUG_MSG0("*** Opened the file.\n");

	for (line = util_getline(fp);line != NULL;line = util_getline(fp)){
		line_number++;
		/*
		 * The format of the gutenfetchrc file is an action word
		 * followed by a single space followed by a string which
		 * depends on the action
		 * ex.
		 * server http://www.ibiblio.org/
		 * verbose true
		 *
		 * This chunk of code makes action the first string 'server'
		 * or 'verbose' and the value string is what follows.
		 */
		value = line;
		action = strsep(&value, " ");
		if (action != NULL) {
			DEBUG_MSG1("*** action: '%s' - ", action);
			if (strcmp(action, "server") == 0) { /* we have a server entry. */
				if (value != NULL) {
					DEBUG_MSG1("%s\n", value);
					gutenerr = gutenfetch_set_active_server( value );
					if( gutenerr != GUTENFETCH_OK ) {
						fprintf(stderr, _("Unable to set active server to %s.\n"),
							value );
					}
				}	
			} else if (strcmp(action, "verbose") == 0) {
				if (value != NULL) {
					DEBUG_MSG1("%s\n", value);
					if ((strcmp(value, "true") == 0) ||	
					(strcmp(value, "1") == 0))
					{
						verbose = TRUE;
					}else if((strcmp(value, "false") == 0) ||
					(strcmp(value, "0") == 0)){
						verbose = FALSE;	
					} else {		
						fprintf(stderr, _("Your configuration file '%s' has a\n\
					        malformed verbose entry on line %u.\n\
							The correct syntax is...\n\n\
							verbose true\t[or]\n\
							verbose false\n"), filename, line_number);
					}		
				}
			} else {
				DEBUG_MSG0("unrecognized\n");
			}	
		}
		FREE_NULL(line);
	}

	if (fclose(fp) != 0) {
		if (verbose) {
			fprintf(stderr, _("Unable to close file '%s'."), filename);
		}
	}	
	DEBUG_MSG0("*** Closed the file.\n");
	return TRUE;
}


/** 
 * Prompt a user to select their default servers.
 *
 * This function allows a user to select their default servers
 * for this session.  It returns true if it was able to select
 * at least one server to use, FALSE otherwise.
 *
 * @return TRUE if the user selected a server(s), FALSE if
 *         no servers were selected by the user.
 */
int
opt_manual_select_servers(void)
{
	//list_t *srv = NULL;
	FILE *std_input = fdopen(0, "r");
	gutenfetch_server_t **pot_servers;
	gutenfetch_server_t *server;
	gutenfetch_error_t gutenerr;

	if (std_input == NULL) {
		fprintf(stderr, _("Unable to get file pointer for standard input.\n"));
		return( FALSE );
	}

	pot_servers = opt_get_continent_selection();
	if( pot_servers == NULL ) {
		fprintf(stderr, _("Unable to determine potential servers.\n") );
		return( FALSE );
	}

	server = opt_get_server_selection( pot_servers );
	gutenerr = gutenfetch_set_active_server_full( server );
	if( gutenerr != GUTENFETCH_OK ) {
		fprintf( stderr, "%s\n", gutenfetch_error_to_string( gutenerr ) );
	}

	gutenfetch_free_server( server );
	gutenfetch_free_servers( pot_servers );
	
	return( TRUE );
}


/**
 * Prompts the user for desired servers and tries to write a new .gutenfetchrc file.
 */
void
opt_generate_new_config_file(void)
{
	if (opt_manual_select_servers() == TRUE) {
		opt_write_config();
	}	
}


/**
 * Select the server to add to the active server list.
 *
 * This is given a list of potential servers.  It displays
 * them to stdout and lets the user select a server.  It
 * returns the URL of the server.  It is intended to be used
 * internally by opt.c.
 *
 * @param srv A list_t of server_t*.  The user can select
 *      one of these servers.
 * @return a URL of the server the user selected.  This value
 *		must be freed when no longer needed.
 */
gutenfetch_server_t * 
opt_get_server_selection(gutenfetch_server_t **srv)
{
	//char *retval;
	int index = 0, key;//, list_len;

	if( srv == NULL )
		return( NULL );

	for(key = -1; (key < 0) || (key >= index); scanf( "%d", &key) ) {
		for(index = 0; srv[index] != NULL; ++index ) {
			fprintf(stderr, "[%u] - %s - %s\n",
				index,
				srv[index]->name,
				srv[index]->area);
		}
		fprintf(stderr, _("Select a server:  "));
	}

	assert( srv[key] != NULL );

	return( gutenfetch_duplicate_server( srv[key] ) );
}


/**
 * Get the continent the user would like to select a server from.
 *
 * @return A list_t* of servers on the continent that the user
 *		selected.
 */
gutenfetch_server_t **
opt_get_continent_selection(void)
{
	int key = -1;
	char *line = NULL;
	FILE *std_input = fdopen(0, "r");
	gutenfetch_server_t **srv = NULL;

	if (std_input == NULL) {
		fprintf(stderr, _("Unable to get file pointer for standard input.\n"));
		return NULL;
	}
	
	while ( (key < 0) || (key > 5) ) {
		fprintf(stderr, "[0] - %s\n", _("North America"));
		fprintf(stderr, "[1] - %s\n", _("South America"));
		fprintf(stderr, "[2] - %s\n", _("Europe"));
		fprintf(stderr, "[3] - %s\n", _("Africa"));
		fprintf(stderr, "[4] - %s\n", _("Austrailia/Oceania"));
		fprintf(stderr, "[5] - %s\n", _("Asia"));
		fprintf(stderr, _("Please choose a continent: "));
		line = util_getline(std_input);
		if (line != NULL) {
			key = (int)strtol(line, NULL, 10);
			FREE_NULL(line);
		} else { // error in getline let's not continue.
			return NULL;
		}
	}	

	if( key == 0 ) {	
		srv = gutenfetch_list_servers( NORTH_AMERICA );
	} else if( key == 1 ) {
		srv = gutenfetch_list_servers( SOUTH_AMERICA );
	} else if( key == 2 ) {
		srv = gutenfetch_list_servers( EUROPE );
	} else if( key == 3 ) {
		srv = gutenfetch_list_servers( AFRICA );
	} else if( key == 4 ) {
		srv = gutenfetch_list_servers( AUSTRALASIA_OCEANIA );
	} else if( key == 5 ) {
		srv = gutenfetch_list_servers( ASIA );
	} else {
		/* This should never occur! */
		abort();
	}

	return srv;
}


/**
 * Get the value of verbose.
 *
 * @return TRUE we should be silent, FALSE if not.
 */
int
opt_get_verbose(void)
{
	return verbose;
}

/**
 * Set the verbosity of gutenfetch.
 *
 * @param v TRUE to be silent, FALSE to be loud.
 */
void
opt_set_verbose(int v)
{
	verbose = v;
}

/**
 * Get the name of the output file.
 *
 * @return The full name of the output file.
 */
const char *
opt_get_output_file(void)
{
	return output_file;
}

/**
 * Set the name of the output file.
 *
 * @param f The filename of the the output file.
 *        or "-" for stdout.
 */
void
opt_set_output_file(const char *f)
{
	FREE_NULL(output_file);
	output_file = strdup(f);
	if ((output_file == NULL) && (verbose == TRUE)) {
		fprintf(stderr, _("Unable to set output file!"));
	}	
}

/**
 * Set the action which we will do.
 * 
 * @param a The action which we will perform.
 */
void
opt_set_action(action_t a)
{
	if ( 		(action != ACTION_NONE) &&
			(action != ACTION_ERROR) &&
			(action != ACTION_FETCH_BOOK) &&
			(action != ACTION_LIST_BOOKS) &&
			(action != ACTION_MANUAL_SERVER_SELECT) &&
			(action != ACTION_PRINT_VERSION) &&
			(action != ACTION_PRINT_USAGE)
	
	) {
		if (verbose) {
			fprintf(stderr, _("Unable to set action, it is invalid."));
		}
	} else	
		action = a;
}

/**
 * Get the action we shall perform.
 *
 * @return The action we will perform.
 */
action_t
opt_get_action(void)
{
	return action;
}


/**
 * Set the bookbase of the book we will fetch.
 *
 * @param b the bookbase of the book we will set.
 * @return TRUE on success, FALSE on failure.
 */
int
opt_set_bookbase(const char *b)
{
	int retval = TRUE;
	
	if ((strlen(b) > 8) || (strlen(b) <= 0)) {
		if (verbose) {
			fprintf(stderr, _("Unable to set bookbase, it is invalid."));
		}
		retval = FALSE;
	} else	{
		bookbase = strdup(b);
		if (bookbase == NULL) {
			if (verbose)
				fprintf(stderr, _("Unable to allocate %u bytes.\n"), strlen(b));
			retval = FALSE;	
		}
	}	
	
	return retval;	
}


/**
 * Get the bookbase of the book we will fetch.
 *
 * @return the bookbase.
 */
const char *
opt_get_bookbase(void)
{
	return bookbase;
}


/**
 * Set the filter flag.
 *
 * @param f TRUE if we should filter output FALSE if not.
 */
void
opt_set_filter(const int f)
{
	filter = f;
}


/**
 * Filter the results 
 * 
 * @return TRUE if we should filter the results before writing.
 */
int
opt_get_filter(void)
{
	return( filter );
}

/**
 * opt_set_cache
 *
 * @param c TRUE if we should cache files, FALSE if not.
 */
void
opt_set_cache(const int c)
{
	enable_cache = c;
}

/**
 * opt_get_cache
 *
 * @return Whether caching is enabled or not.
 */
int
opt_get_cache(void)
{
	return( enable_cache );
}
