/***********************************************************************
	gutenfetch - query and fetch electronic texts from project gutenberg
    Copyright (C) 2001, 2002, 2003 Russell Francis 

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
	USA

	Last updated on $Date: 2004/07/09 02:58:24 $ by $Author: johntabularasa $.
***********************************************************************/
#include "stddefs.h"
#if (HAVE_ERRNO_H == 1)
#	include <errno.h>
#endif
#if (HAVE_STDARG_H == 1)
#	include <stdarg.h>
#endif
#if (HAVE_STDIO_H == 1)
#	include <stdio.h>
#endif
#if (HAVE_STDLIB_H == 1)
#	include <stdlib.h>
#endif
#if (HAVE_STRING_H == 1)
#	include <string.h>
#else
#	if (HAVE_STRINGS_H == 1)
#		include <strings.h>
#	endif
#endif
#if (HAVE_ASSERT_H == 1)
#	include <assert.h>
#endif
#if (HAVE_SYS_TYPES_H == 1)
#	include <sys/types.h>
#endif
#if (HAVE_SYS_UIO_H == 1)
#	include <sys/uio.h>
#endif
#include "util.h"
#include "opt.h"
#include "list.h"

/**
 * util_strcat concatenates strings into one large string.
 *
 * This function takes a NULL terminated list of NULL terminated
 * strings and concatenates them together.
 *
 * @param a A NULL terminated string.
 * @return A string which must be freed when finished.  It will
 *      be either the concatenation of all the strings passed in
 *  	or will be NULL.
 */
char *
util_strcat(char *a, ...)
{
	va_list ap;
	size_t len = 1; /* space for null terminator */
	list_t *l = NULL;
	list_t *head = NULL;
	char *item;

	assert (a != NULL);
	len += strlen(a);
	l = list_append(l, a);

	va_start(ap, a);
	while ( (item = va_arg(ap, char *)) != NULL ) {
		len += strlen(item);
		l = list_append(l, item);
	}
	va_end(ap);
	item = malloc(sizeof(char) * len);
	if ( item == NULL ) {
		fprintf(stderr, _("Unable to allocate %u bytes of memory."),
			(sizeof(char) * len));
		abort();
	}
	
	head = l = list_first(l);
	strcpy(item, l->data);

	for(l = list_next(l); l != NULL ; l = list_next(l))
		strcat(item, l->data);
	
	list_remove_all(head, NULL);
	return item;
}


/**
 * Get the next line from a file.
 *
 * This function returns the characters in a file up to the '\n'
 * character.  The returned result is a NULL terminated string.
 *
 * @param fp The file pointer of the file to read from.
 * @return a NULL terminated string which contains the next line
 * 			in the file.
 */
char *
util_getline(FILE *fp)
{
	char *line = NULL;
	char *temp = NULL;
	size_t total_size;
	size_t i;

	for (i = 0, total_size = 0; TRUE ; ++i) {
		/* allocate more memory for the line if needed. */
		if(i == total_size){
			total_size += BLOCK_SIZE;
			temp = realloc(line, sizeof(char) * total_size);
			if (temp == NULL) {
				FREE_NULL(line);
				if (opt_get_verbose()) {
					fprintf(stderr, 
					_("Unable to allocate %u bytes of memory."),
					sizeof(char) * total_size);
				}
				break;
			}
			line = temp;
		}

		/* Read the next character into the line */
		if((line[i] = fgetc(fp)) == '\n'){
			line[i] = '\0';	
			break;
		}else if(line[i] == EOF){
			if(i == 0) {
				FREE_NULL(line);
			} else {
				line[i] = '\0';
			}	
			break;
		}
	}
	return line;
}

/**
 * Get a users home directory. 
 *
 * gethomedir() returns the $HOME directory as specified
 * by the environment variable HOME=
 *
 * @return The home directory as a NULL terminated string.
 *         This value should not be freed or modified.
 */
char *
util_gethomedir(void)
{
#ifdef WIN32
	/* Currently, we write a configuration file in the
	   current working directory if we are on windows.
	   This is wrong and a bit of a hack I think, perhaps
	   things should go in the registry?  I really am not
	   sure, if some Windows guru know, feel free to let 
	   me know.
	*/   
#ifndef PATH_MAX
#	define PATH_MAX 256
#endif /* PATH_MAX */
	char hdir[PATH_MAX];
	char *home_dir = NULL;
	home_dir = getcwd(home_dir, PATH_MAX);
#else	
	char *home_dir = getenv("HOME");
#endif	

	if (home_dir == NULL) {
		if (opt_get_verbose()) {
			fprintf(stderr, 
				_("Unable to find the environment variable 'HOME'!"));
		}	
	}
	
	return home_dir;
}

#ifndef HAVE_STRSEP
/**
 *   The strsep() function locates, in the string referenced by *stringp, the
 *   first occurrence of any character in the string delim (or the terminating
 *   `\0' character) and replaces it with a `\0'.  The location of the next
 *   character after the delimiter character (or NULL, if the end of the
 *   string was reached) is stored in *stringp.  The original value of
 *   *stringp is returned.
 * 
 *   An ``empty'' field (i.e., a character in the string delim occurs as the
 *   first character of *stringp) can be detected by comparing the location
 *   referenced by the returned pointer to `\0'.
 *                                                                              
 *   If *stringp is initially NULL, strsep() returns NULL.
 *
 * @param stringp The string to separate.
 * @param delim The character to separate the string stringp on.
 * @return NULL or the next character after the first found delim.
 */
char *
strsep(char **stringp, const char *delim)
{
	char *res;

	if (!stringp || !*stringp || !**stringp)
		return NULL;
	
	res = *stringp;
	while (**stringp && !strchr(delim, **stringp))
		++(*stringp);
	
	if (**stringp) {
		**stringp = '\0';
		++(*stringp);
	}

	return res;
}
#endif /* HAVE_STRSEP */


#ifndef HAVE_MERGESORT
/*-
 * Copyright (c) 1992, 1993
 *	The Regents of the University of California.  All rights reserved.
 *
 * This code is derived from software contributed to Berkeley by
 * Peter McIlroy.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by the University of
 *	California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */
/*
 * The following mergesort function and supporting functions were taken
 * from FreeBSD to provide a sorting routine on machines which do not 
 * provide mergesort().  
 *
 * Russell Francis <johntabularasa@users.sf.net> 11-15-03
 */
#ifndef u_char
#	define u_char unsigned char
#endif
#define ISIZE sizeof(int)
#define PSIZE sizeof(u_char *)
#define ICOPY_LIST(src, dst, last)				\
	do							\
	*(int*)dst = *(int*)src, src += ISIZE, dst += ISIZE;	\
	while(src < last)
#define ICOPY_ELT(src, dst, i)					\
	do							\
	*(int*) dst = *(int*) src, src += ISIZE, dst += ISIZE;	\
	while (i -= ISIZE)

#define CCOPY_LIST(src, dst, last)		\
	do					\
		*dst++ = *src++;		\
	while (src < last)
#define CCOPY_ELT(src, dst, i)			\
	do					\
		*dst++ = *src++;		\
	while (i -= 1)

/*
 * Find the next possible pointer head.  (Trickery for forcing an array
 * to do double duty as a linked list when objects do not align with word
 * boundaries.
 */
/* Assumption: PSIZE is a power of 2. */
#define EVAL(p) (u_char **)						\
	((u_char *)0 +							\
	    (((u_char *)p + PSIZE - 1 - (u_char *) 0) & ~(PSIZE - 1)))

/*
 * Arguments are as for qsort.
 */
int
mergesort(base, nmemb, size, cmp)
	void *base;
	size_t nmemb;
	size_t size;
	int (*cmp)(const void *, const void *);
{
	int i, sense;
	int big, iflag;
	u_char *f1, *f2, *t, *b, *tp2, *q, *l1, *l2;
	u_char *list2, *list1, *p2, *p, *last, **p1;

	if (size < PSIZE / 2) {		/* Pointers must fit into 2 * size. */
		errno = EINVAL;
		return (-1);
	}

	if (nmemb == 0)
		return (0);

	/*
	 * XXX
	 * Stupid subtraction for the Cray.
	 */
	iflag = 0;
	if (!(size % ISIZE) && !(((char *)base - (char *)0) % ISIZE))
		iflag = 1;

	if ((list2 = malloc(nmemb * size + PSIZE)) == NULL)
		return (-1);

	list1 = base;
	mergesort_setup(list1, list2, nmemb, size, cmp);
	last = list2 + nmemb * size;
	i = big = 0;
	while (*EVAL(list2) != last) {
	    l2 = list1;
	    p1 = EVAL(list1);
	    for (tp2 = p2 = list2; p2 != last; p1 = EVAL(l2)) {
	    	p2 = *EVAL(p2);
	    	f1 = l2;
	    	f2 = l1 = list1 + (p2 - list2);
	    	if (p2 != last)
	    		p2 = *EVAL(p2);
	    	l2 = list1 + (p2 - list2);
	    	while (f1 < l1 && f2 < l2) {
	    		if ((*cmp)(f1, f2) <= 0) {
	    			q = f2;
	    			b = f1, t = l1;
	    			sense = -1;
	    		} else {
	    			q = f1;
	    			b = f2, t = l2;
	    			sense = 0;
	    		}
	    		if (!big) {	/* here i = 0 */
				while ((b += size) < t && cmp(q, b) >sense)
	    				if (++i == 6) {
	    					big = 1;
	    					goto EXPONENTIAL;
	    				}
	    		} else {
EXPONENTIAL:	    		for (i = size; ; i <<= 1)
	    				if ((p = (b + i)) >= t) {
	    					if ((p = t - size) > b &&
						    (*cmp)(q, p) <= sense)
	    						t = p;
	    					else
	    						b = p;
	    					break;
	    				} else if ((*cmp)(q, p) <= sense) {
	    					t = p;
	    					if (i == size)
	    						big = 0;
	    					goto FASTCASE;
	    				} else
	    					b = p;
				while (t > b+size) {
	    				i = (((t - b) / size) >> 1) * size;
	    				if ((*cmp)(q, p = b + i) <= sense)
	    					t = p;
	    				else
	    					b = p;
	    			}
	    			goto COPY;
FASTCASE:	    		while (i > size)
	    				if ((*cmp)(q,
	    					p = b + (i >>= 1)) <= sense)
	    					t = p;
	    				else
	    					b = p;
COPY:	    			b = t;
	    		}
	    		i = size;
	    		if (q == f1) {
	    			if (iflag) {
	    				ICOPY_LIST(f2, tp2, b);
	    				ICOPY_ELT(f1, tp2, i);
	    			} else {
	    				CCOPY_LIST(f2, tp2, b);
	    				CCOPY_ELT(f1, tp2, i);
	    			}
	    		} else {
	    			if (iflag) {
	    				ICOPY_LIST(f1, tp2, b);
	    				ICOPY_ELT(f2, tp2, i);
	    			} else {
	    				CCOPY_LIST(f1, tp2, b);
	    				CCOPY_ELT(f2, tp2, i);
	    			}
	    		}
	    	}
	    	if (f2 < l2) {
	    		if (iflag)
	    			ICOPY_LIST(f2, tp2, l2);
	    		else
	    			CCOPY_LIST(f2, tp2, l2);
	    	} else if (f1 < l1) {
	    		if (iflag)
	    			ICOPY_LIST(f1, tp2, l1);
	    		else
	    			CCOPY_LIST(f1, tp2, l1);
	    	}
	    	*p1 = l2;
	    }
	    tp2 = list1;	/* swap list1, list2 */
	    list1 = list2;
	    list2 = tp2;
	    last = list2 + nmemb*size;
	}
	if (base == list2) {
		memmove(list2, list1, nmemb*size);
		list2 = list1;
	}
	free(list2);
	return (0);
}

#define	swap(a, b) {					\
		s = b;					\
		i = size;				\
		do {					\
			tmp = *a; *a++ = *s; *s++ = tmp; \
		} while (--i);				\
		a -= size;				\
	}
#define reverse(bot, top) {				\
	s = top;					\
	do {						\
		i = size;				\
		do {					\
			tmp = *bot; *bot++ = *s; *s++ = tmp; \
		} while (--i);				\
		s -= size2;				\
	} while(bot < s);				\
}

/*
 * Optional hybrid natural/pairwise first pass.  Eats up list1 in runs of
 * increasing order, list2 in a corresponding linked list.  Checks for runs
 * when THRESHOLD/2 pairs compare with same sense.  (Only used when NATURAL
 * is defined.  Otherwise simple pairwise merging is used.)
 */
void
mergesort_setup(list1, list2, n, size, cmp)
	size_t n, size;
	int (*cmp)(const void *, const void *);
	u_char *list1, *list2;
{
	int i, length, size2, tmp, sense;
	u_char *f1, *f2, *s, *l2, *last, *p2;

	size2 = size*2;
	if (n <= 5) {
		mergesort_insertionsort(list1, n, size, cmp);
		*EVAL(list2) = (u_char*) list2 + n*size;
		return;
	}
	/*
	 * Avoid running pointers out of bounds; limit n to evens
	 * for simplicity.
	 */
	i = 4 + (n & 1);
	mergesort_insertionsort(list1 + (n - i) * size, i, size, cmp);
	last = list1 + size * (n - i);
	*EVAL(list2 + (last - list1)) = list2 + n * size;

#ifdef NATURAL
	p2 = list2;
	f1 = list1;
	sense = (cmp(f1, f1 + size) > 0);
	for (; f1 < last; sense = !sense) {
		length = 2;
					/* Find pairs with same sense. */
		for (f2 = f1 + size2; f2 < last; f2 += size2) {
			if ((cmp(f2, f2+ size) > 0) != sense)
				break;
			length += 2;
		}
		if (length < THRESHOLD) {		/* Pairwise merge */
			do {
				p2 = *EVAL(p2) = f1 + size2 - list1 + list2;
				if (sense > 0)
					swap (f1, f1 + size);
			} while ((f1 += size2) < f2);
		} else {				/* Natural merge */
			l2 = f2;
			for (f2 = f1 + size2; f2 < l2; f2 += size2) {
				if ((cmp(f2-size, f2) > 0) != sense) {
					p2 = *EVAL(p2) = f2 - list1 + list2;
					if (sense > 0)
						reverse(f1, f2-size);
					f1 = f2;
				}
			}
			if (sense > 0)
				reverse (f1, f2-size);
			f1 = f2;
			if (f2 < last || cmp(f2 - size, f2) > 0)
				p2 = *EVAL(p2) = f2 - list1 + list2;
			else
				p2 = *EVAL(p2) = list2 + n*size;
		}
	}
#else		/* pairwise merge only. */
	for (f1 = list1, p2 = list2; f1 < last; f1 += size2) {
		p2 = *EVAL(p2) = p2 + size2;
		if (cmp (f1, f1 + size) > 0)
			swap(f1, f1 + size);
	}
#endif /* NATURAL */
}

/*
 * This is to avoid out-of-bounds addresses in sorting the
 * last 4 elements.
 */
static void
mergesort_insertionsort(a, n, size, cmp)
	u_char *a;
	size_t n, size;
	int (*cmp)(const void *, const void *);
{
	u_char *ai, *s, *t, *u, tmp;
	int i;

	for (ai = a+size; --n >= 1; ai += size)
		for (t = ai; t > a; t -= size) {
			u = t - size;
			if (cmp(u, t) <= 0)
				break;
			swap(u, t);
		}
}
#endif /* HAVE_MERGESORT */
